from dataclasses import dataclass
from datetime import datetime, timedelta, time as time_class
from enum import IntEnum
from typing import Set, Union


class Day(IntEnum):
    MONDAY = 0
    TUESDAY = 1
    WEDNESDAY = 2
    THURSDAY = 3
    FRIDAY = 4
    SATURDAY = 5
    SUNDAY = 6

    @classmethod
    def today(cls) -> 'Day':
        today = datetime.now()
        return cls(today.weekday())

    def __sub__(self, other):
        if not isinstance(other, Day):
            raise TypeError(f'Wrong type provided, got {other}')
        return self.value - other.value


class ScheduleConfigurationError(Exception):
    pass


@dataclass
class Instant(object):
    day: Day
    time: time_class

    @classmethod
    def now(cls) -> 'Instant':
        now = datetime.now()
        day = Day(now.weekday())
        time = now.time()
        return Instant(day, time)

    def next_datetime(self) -> datetime:
        now = datetime.now()
        days_ahead = self.day - Day.today()

        is_in_the_past = days_ahead < 0 or (days_ahead == 0 and now.time() > self.time)
        if is_in_the_past:
            days_ahead += 7

        next_date = now + timedelta(days_ahead)
        future_datetime = datetime.combine(next_date.date(), self.time)
        return future_datetime


class Schedule(object):

    def __init__(self):
        self.days: Set[Day] = set()
        self.time: Union[time_class, None] = None

    def __repr__(self):
        return f'{self.__class__.__qualname__} on {self.days} at {self.time}'

    def instants(self):
        self._raise_exception_if_not_valid()
        for day in self.days:
            yield Instant(day, self.time)

    def at(self, time_string: str):
        self.time = self._time_string_to_datetime(time_string)
        return self

    @classmethod
    def _time_string_to_datetime(cls, time_string: str):
        split_string = cls._split_time_string(time_string)
        time_format = {1: '%H', 2: '%H:%M', 3: '%H:%M:%S'}
        time_format_string = time_format[len(split_string)]

        schedule_datetime = datetime.strptime(time_string, time_format_string)
        return schedule_datetime.time()

    @staticmethod
    def _split_time_string(time_string: str):
        split_string = time_string.split(':')
        if 1 < len(split_string) > 3:  # valid 1 <= x <= 3
            e = f'Invalid time string provided, {time_string}'
            raise ScheduleConfigurationError(e)
        return split_string

    def every(self, *days) -> 'Schedule':
        """
        Sets the schedule to be active on the specified days.
        @param days: the exact days when the schedule will be active
        """
        self.days = set(days)
        return self

    def every_day(self) -> 'Schedule':
        """
        Sets the schedule to be active every day of the week.
        """
        self.days = {day for day in Day}
        return self

    def exclude(self, *days) -> 'Schedule':
        """
        Excludes the specified days from the schedule.

        Example:
        >>> schedule = Schedule().between(Day.Monday, Day.FRIDAY)
        >>> schedule.exclude(Day.THURSDAY)
        >>> print(schedule)
        @param days: the days to remove from the schedule
        """
        excluded_days = set(days)
        self.days = self.days - excluded_days
        return self

    def between(self, *days) -> 'Schedule':
        """
        TODO: this might be confusing with time; consider if it should be removed.
        Sets the schedule to be active between the specified days (inclusive).
        @param days: the extremes of the interval
        """
        if len(days) != 2:
            e = f'Schedules between days require exactly two days, got {days}'
            raise ScheduleConfigurationError(e)

        start_day, end_day = days[0], days[1]
        self.days = self._weekdays_between(start_day, end_day)
        return self

    @staticmethod
    def _weekdays_between(start_day: Day, end_day: Day) -> Set[Day]:
        if start_day == end_day:  # e.g. Friday to Friday
            end_day += 6
        if start_day > end_day:  # e.g. Saturday to Tuesday
            end_day += 7

        days_interval = {Day(index % 7) for index in range(start_day, end_day + 1)}
        return days_interval

    def weekdays(self):
        """Monday to Friday"""
        self.days = {day for day in Day}
        self.exclude(Day.SATURDAY, Day.SUNDAY)
        return self

    def weekends(self):
        """Saturday and Sunday"""
        self.days = {Day.SATURDAY, Day.SUNDAY}
        return self

    def is_active_today(self) -> bool:
        """
        Returns true if the schedule is set to be active today.
        """
        self._raise_exception_if_not_valid()

        today = datetime.now()
        day = Day(today.weekday())
        return day in self.days

    def is_valid(self) -> bool:
        return len(self.days) != 0 and self.time is not None

    def _raise_exception_if_not_valid(self) -> None:
        if not self.is_valid():
            e = 'Valid schedules must specify a time and at least one day.'
            raise ScheduleConfigurationError(e)



import pytest
from unittest import TestCase

from symbiotic.devices import LightBulb
from symbiotic.services import BaseService, ServiceResponse

SERVICE_CALL_SUCCESS = 'mock-success-call'
SERVICE_CALL_FAIL = 'mock-fail-call'


###########################################
#  ______ _      _
# |  ____(_)    | |
# | |__   ___  _| |_ _   _ _ __ ___  ___
# |  __| | \ \/ / __| | | | '__/ _ \/ __|
# | |    | |>  <| |_| |_| | | |  __/\__ \
# |_|    |_/_/\_\\__|\__,_|_|  \___||___/
###########################################


@pytest.fixture(scope='function')
def service_success(request) -> BaseService:
    class ServiceSuccess(BaseService):
        def trigger(self, *args, **kwargs):
            return ServiceResponse(
                success=True,
                message=SERVICE_CALL_SUCCESS,
            )

    request.cls.service = ServiceSuccess()


@pytest.fixture(scope='function')
def service_fail(request) -> BaseService:
    class ServiceFail(BaseService):
        def trigger(self, *args, **kwargs):
            return ServiceResponse(
                success=False,
                message=SERVICE_CALL_FAIL,
            )

    request.cls.service = ServiceFail()


#############################
#   _______        _
#  |__   __|      | |
#     | | ___  ___| |_ ___
#     | |/ _ \/ __| __/ __|
#     | |  __/\__ \ |_\__ \
#     |_|\___||___/\__|___/
#############################


@pytest.mark.usefixtures('service_success')
class Test_Unit_LightBulb(TestCase):
    def test_create_light_bulb(self) -> None:
        light_bulb = LightBulb('room', service=self.service)
        self.assertIsNotNone(light_bulb)
        self.assertEqual(type(light_bulb), LightBulb)


@pytest.mark.usefixtures('service_success')
class Test_Integration_LightBulb_Success(TestCase):

    def test_light_bulb_turn_on_success(self) -> None:
        light_bulb = LightBulb('room', service=self.service)
        response = light_bulb.turn_on()
        self.assertTrue(response.success)
        self.assertEqual(response.message, SERVICE_CALL_SUCCESS)

    def test_light_bulb_turn_off_success(self) -> None:
        light_bulb = LightBulb('room', service=self.service)
        response = light_bulb.turn_off()
        self.assertTrue(response.success)
        self.assertEqual(response.message, SERVICE_CALL_SUCCESS)


@pytest.mark.usefixtures('service_fail')
class Test_Integration_LightBulb_Fail(TestCase):

    def test_light_bulb_turn_off_fail(self) -> None:
        light_bulb = LightBulb('room', service=self.service)
        response = light_bulb.turn_off()
        self.assertFalse(response.success)
        self.assertEqual(response.message, SERVICE_CALL_FAIL)

    def test_light_bulb_turn_on_fail(self) -> None:
        light_bulb = LightBulb('room', service=self.service)
        response = light_bulb.turn_on()
        self.assertFalse(response.success)
        self.assertEqual(response.message, SERVICE_CALL_FAIL)



from datetime import datetime
from functools import partial
from typing import Callable, List, Union

from symbiotic.schedule import Schedule


class Action(object):

    def __init__(self, callback: Callable, *args, **kwargs):
        self._callback: partial = partial(callback, *args, **kwargs)
        self._schedule: Union[Schedule, None] = None
        self._next_execution: Union[datetime, None] = None

    def __repr__(self):
        rep = f'{self.__class__.__qualname__}:'
        rep += f' {self._callback.func.__name__},'
        rep += f' args: {self._callback.args},'
        rep += f' kwargs: {self._callback.keywords}'
        return rep

    def __call__(self):
        return self._callback()

    def set_schedule(self, schedule: Schedule) -> None:
        self._schedule = schedule
        self.schedule_next_execution()

    def should_execute(self):
        return datetime.now() > self._next_execution

    def schedule_next_execution(self):
        datetimes = [instant.next_datetime() for instant in self._schedule.instants()]
        self._next_execution = min(datetimes)  # get the earliest execution datetime


class ActionScheduler(object):

    def __init__(self):
        self.actions: List[Action] = []
        self._schedule: Union[Schedule, None] = None

    def start_session(self, schedule: Schedule):
        self._schedule = schedule

    def add(self, callback: Callable, *args, **kwargs):
        action = Action(callback, *args, *kwargs)
        action.set_schedule(self._schedule)
        self.actions.append(action)
        return action

    def end_session(self):
        self._schedule = None

    def run(self):
        for action in self.actions[:]:
            if action.should_execute():
                action()
                action.schedule_next_execution()



import functools
from abc import ABC, abstractmethod
from collections import Callable

from event_bus import EventBus as SimpleEventBus


class EventBus(ABC):

    @abstractmethod
    def subscribe_func_to_event(self, func: Callable, event_name: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def unsubscribe_func_from_event(self, func: Callable, event_name: str) -> None:
        raise NotImplementedError

    def subscribe(self, event_name: str) -> Callable:
        def event_wrapper(func):
            self.subscribe_func_to_event(func, event_name)

            @functools.wraps(func)
            def func_wrapper(*args, **kwargs):
                return func(*args, **kwargs)

            return func_wrapper

        return event_wrapper

    @abstractmethod
    def emit(self, event_name: str, *args, **kwargs) -> None:
        raise NotImplementedError


class EventBusAdapter(EventBus):

    def __init__(self):
        self._bus = SimpleEventBus()

    def subscribe_func_to_event(self, func: Callable, event_name: str) -> None:
        self._bus.add_event(func, event_name)

    def unsubscribe_func_from_event(self, func: Callable, event_name: str) -> None:
        self._bus.remove_event(func.__name__, event_name)

    def emit(self, event_name: str, *args, **kwargs) -> None:
        self._bus.emit(event_name, *args, **kwargs)


class EventSubscriber(object):

    def __init__(self, event_bus: EventBus, event_name: str):
        self.event_bus = event_bus
        self.event_name = event_name

    def do(self, func: Callable, *args, **kwargs):
        subscriber = functools.partial(func, *args, **kwargs)
        self.event_bus.subscribe_func_to_event(subscriber, self.event_name)



import atexit
import logging
import sys
import time
from contextlib import contextmanager

from dependency_injector import providers
from dependency_injector.providers import Configuration

from .actions import ActionScheduler
from .containers import (
    Container,
    DeviceContainer,
    SensorContainer,
    ServiceContainer,
)
from .event_bus import EventBusAdapter, EventSubscriber
from .schedule import Schedule


class Symbiotic(object):

    def __init__(self):
        self.container: Container = self.create_container()
        self._scheduler: ActionScheduler = ActionScheduler()
        atexit.register(self.shutdown)

    def create_container(self) -> Container:
        container = Container(event_bus=providers.Singleton(EventBusAdapter))
        container.config.debug.from_env('SYMBIOTIC_DEBUG')
        container.init_resources()
        container.wire(modules=[sys.modules[__name__]])
        return container

    @property
    def config(self) -> Configuration:
        return self.container.config

    @property
    def debug(self) -> bool:
        debug = self.config.debug
        if debug is not None:
            return debug
        return False

    @property
    def name(self) -> str:
        return 'Symbiotic'

    @property
    def logger(self) -> logging.Logger:
        logger = logging.getLogger(self.name)

        if self.debug and not logger.level:
            logger.setLevel(logging.DEBUG)

        return logger

    @property
    def devices(self) -> DeviceContainer:
        return self.container.devices

    @contextmanager
    def events(self, event_name: str) -> EventSubscriber:
        event_subscriber = EventSubscriber(self.event_bus, event_name)
        yield event_subscriber

    @property
    def event_bus(self):
        return self.container.event_bus()

    @contextmanager
    def scheduler(self, schedule: Schedule):
        self._scheduler.start_session(schedule)
        yield self._scheduler
        self._scheduler.end_session()

    @property
    def sensors(self) -> SensorContainer:
        return self.container.sensors

    @property
    def services(self) -> ServiceContainer:
        return self.container.services

    def run(self, sleep_interval: int = 1) -> None:
        try:
            self.logger.info(
                'The application is running... (Press CTRL+C to terminate)')
            while True:
                self._scheduler.run()
                time.sleep(sleep_interval)
        except KeyboardInterrupt:
            pass

    def shutdown(self, *args) -> None:
        # sys.stderr.write("\r")  # suppress '^C' in terminal
        # https://stackoverflow.com/a/48726537/5874339
        self.logger.info('Shutdown initiated. Please wait...')
        # Handle application shutdown here...
        self.container.shutdown_resources()
        self.logger.info('Application successfully shutdown.')
        sys.exit(0)



import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any

import requests
from schema import And, Optional, Or, Schema

from symbiotic.exceptions import ConfigurationError
from symbiotic.parameters import Parameters


@dataclass
class ServiceResponse(object):
    success: bool
    message: str

    @staticmethod
    def from_response(response: requests.Response):
        success = response.ok
        return ServiceResponse(success=success, message=response.text)


class BaseService(ABC):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def __repr__(self):
        return f'{self.__class__.__name__}'

    @abstractmethod
    def trigger(self, *args, **kwargs) -> ServiceResponse:
        pass


class IFTTT(BaseService):

    DEFAULT_URL = 'https://maker.ifttt.com/trigger/{event_name}/with/key/{key}'

    def __init__(self, config: dict, *args, **kwargs):
        super().__init__(*args, **kwargs)
        try:
            self._url = config.get('url', IFTTT.DEFAULT_URL)
            self._key = config.pop('key')
        except (AttributeError, KeyError):
            raise ConfigurationError(
                'Could not instantiate IFTTT: configuration not found!')

    def _validate_parameters(self, parameters: Any = None) -> dict:
        if parameters is None:
            return {}

        if issubclass(type(parameters), Parameters):
            parameters = parameters.ifttt()

        schema = Schema({
            Optional('value1'): And(Or(int, float, str)),
            Optional('value2'): And(Or(int, float, str)),
            Optional('value3'): And(Or(int, float, str)),
        })

        return schema.validate(parameters)

    def trigger(self, event_name: str, parameters: Any = None) -> ServiceResponse:
        """Triggers the IFTTT webhook 'event_name' with 'parameters'.

        Parameters
        ----------
        event_name : str
            The name of the webhook to trigger, as defined in the "Event Name".

        parameters : dict, optional
            An optional dictionary containing the parameters to pass with the request.
            It can contain one to three parameters defined as follows:
            {
                (Optional) 'value1': '...',
                (Optional) 'value2': '...',
                (Optional) 'value3': '...'
            }

        """
        parameters = self._validate_parameters(parameters)
        url = self._url.format(event_name=event_name, key=self._key)
        response = requests.post(url, parameters)

        logging.debug(f'Request body: {response.request.body}')
        logging.info(f'{response.text}')

        print(response)

        return ServiceResponse.from_response(response)



from typing import Set, Callable
from unittest import TestCase, mock

import io
import pytest

from symbiotic.event_bus import EventBusAdapter

EVENT_NAME: str = 'test-event'


@pytest.fixture
def mock_function():
    pass


class Test_EventBus(TestCase):

    def setUp(self) -> None:
        self.event_bus: EventBusAdapter = EventBusAdapter()
        self.assertIsNotNone(self.event_bus)

    def event_subscribers(self, event_name) -> Set[Callable]:
        return self.event_bus._bus._events[event_name]

    @pytest.mark.usefixtures('mock_function')
    def test_subscribe_func_to_event(self):
        self.event_bus.subscribe_func_to_event(mock_function, EVENT_NAME)
        event_subscribers = self.event_subscribers(EVENT_NAME)
        expected: set = {mock_function}
        self.assertEqual(expected, event_subscribers)

    @pytest.mark.usefixtures('mock_function')
    def test_unsubscribe_func_from_event(self):
        self.event_bus.subscribe_func_to_event(mock_function, EVENT_NAME)
        self.event_bus.unsubscribe_func_from_event(mock_function, EVENT_NAME)
        self.assertEqual(set(), self.event_subscribers(EVENT_NAME))

    def test_emit_event(self):
        mock_func_result = 'mock_function called'
        expected_result = 'mock_function called\n'

        def mock_func_print():
            print(mock_func_result)

        with mock.patch('sys.stdout', new_callable=io.StringIO) as mock_stdout:
            self.event_bus.subscribe_func_to_event(mock_func_print, EVENT_NAME)
            self.event_bus.emit(EVENT_NAME)
            self.assertEqual(expected_result, mock_stdout.getvalue())



import logging
from abc import ABC

from gpiozero import MotionSensor as GPIOZeroMotionSensor
from gpiozero.exc import BadPinFactory

from .event_bus import EventBus
from .exceptions import ConfigurationError


class MotionSensor(ABC):
    """
    MotionSensor provides an interface for motion sensors.

    When motion is detected, the sensor should call `active`;
    this will emit an event on the bus as `sensor_name:active`.

    Args:
        name (str): the name to associate with the motion sensor.
    """

    def __init__(self, name: str, *args, **kwargs):
        self.name: str = name
        self.bus: EventBus = kwargs.pop('event_bus')
        super().__init__(*args, **kwargs)

    @property
    def movement_detected(self) -> str:
        return f'{self.name}:active'

    @property
    def movement_stopped(self) -> str:
        return f'{self.name}:inactive'

    def _movement_detected_hook(self):
        logging.debug(f'{self.name}: movement detected.')
        self.bus.emit(self.movement_detected)

    def _movement_stopped_hook(self):
        logging.debug(f'{self.name}: movement stopped.')
        self.bus.emit(self.movement_stopped)


class GPIOMotionSensor(MotionSensor):

    def __init__(self, name: str, pin: int, *args, **kwargs):
        pin_factory = kwargs.pop('pin_factory', None)
        self._initialise(pin, pin_factory)
        self._bind_actions()
        super().__init__(name, *args, **kwargs)

    def _initialise(self, pin: int, pin_factory=None) -> None:
        try:
            self._sensor = GPIOZeroMotionSensor(pin, pin_factory=pin_factory)
        except BadPinFactory:
            err = f'Cannot instantiate {self.name} without a pin factory!'
            raise ConfigurationError(err)

    def _bind_actions(self) -> None:
        self._sensor.when_motion = self._movement_detected_hook
        self._sensor.when_no_motion = self._movement_stopped_hook



from enum import Enum
from unittest import TestCase, mock

import pytest
import schema
from symbiotic.services import IFTTT


class StatusCodes(Enum):
    # pylint: disable=unsubscriptable-object
    # the above to suppress errors for the instance methods
    # https://github.com/PyCQA/pylint/issues/2063

    OK = (200, 'ok')
    NO_CONTENT = (204, 'no_content')
    BAD_REQUEST = (400, 'bad_request')
    UNAUTHORIZED = (401, 'unauthorized')
    FORBIDDEN = (403, 'forbidden')
    NOT_FOUND = (404, 'not_found')
    INVALID_PARAMETERS = (422, 'invalid_parameters')
    INTERNAL_SERVER_ERROR = (500, 'internal_server_error')
    BAD_GATEWAY = (502, 'bad_gateway')

    @staticmethod
    def by_reason(reason: str) -> 'StatusCodes':
        for item in StatusCodes:
            if reason == item.value[1]:
                return item
        raise Exception(f'Reason not found in status codes: {reason}')

    @staticmethod
    def has_reason(reason: str) -> str:
        values = set(item.value[1] for item in StatusCodes)
        return reason in values

    @property
    def code(self) -> int:
        return self.value[0]

    @property
    def reason(self) -> str:
        return self.value[1]


def ifttt_service_valid_key() -> IFTTT:
    config = {'key': 'clearly_a_valid_key'}
    return IFTTT(config=config)


def ifttt_service_invalid_key() -> IFTTT:
    config = {'key': 'clearly_an_invalid_key'}
    return IFTTT(config=config)


@mock.patch('symbiotic.services.requests.post', autospec=True)
class Test_IFTTT_Unit(TestCase):

    def test_trigger_valid_request_no_params(self, mock_post):
        ifttt = IFTTT(config={'key': 'valid_key'})
        response = ifttt.trigger(event_name='name')
        self.assertTrue(response.success)

    def test_trigger_valid_request_with_params1(self, mock_post):
        ifttt = IFTTT(config={'key': 'valid_key'})
        response = ifttt.trigger(event_name='name', parameters={'value1': 42})
        self.assertTrue(response.success)

    def test_trigger_valid_request_with_params2(self, mock_post):
        ifttt = IFTTT(config={'key': 'valid_key'})
        response = ifttt.trigger(event_name='name', parameters={'value2': 42, 'value3': 'some-value'})
        self.assertTrue(response.success)

    def test_trigger_valid_request_with_params3(self, mock_post):
        ifttt = IFTTT(config={'key': 'valid_key'})
        response = ifttt.trigger(event_name='name', parameters={})
        self.assertTrue(response.success)

    def test_trigger_valid_request_with_params4(self, mock_post):
        ifttt = IFTTT(config={'key': 'valid_key'})
        response = ifttt.trigger(event_name='name', parameters={})
        self.assertTrue(response.success)

    def test_trigger_valid_request_with_params5(self, mock_post):
        ifttt = IFTTT(config={'key': 'valid_key'})
        with pytest.raises(schema.SchemaUnexpectedTypeError):
            ifttt.trigger(event_name='name', parameters='test')

    def test_trigger_valid_request_with_params6(self, mock_post):
        ifttt = IFTTT(config={'key': 'valid_key'})
        with pytest.raises(TypeError):
            ifttt.trigger(event_name='name', color='black')

    def test_trigger_valid_request_with_params7(self, mock_post):
        ifttt = IFTTT(config={'key': 'valid_key'})
        with pytest.raises(schema.SchemaWrongKeyError):
            ifttt.trigger(event_name='name', parameters={'value1': 'some-value', 'value4': 55})

    def test_trigger_valid_request_with_params8(self, mock_post):
        ifttt = IFTTT(config={'key': 'valid_key'})
        with pytest.raises(schema.SchemaWrongKeyError):
            ifttt.trigger(event_name='name', parameters={'value1': 'some-value', 'sheep': 'baaah'})



"""Http client module."""

from aiohttp import ClientSession, ClientTimeout, ClientResponse


class HttpClient:

    async def request(self, method: str, url: str, timeout: int) -> ClientResponse:
        async with ClientSession(timeout=ClientTimeout(timeout)) as session:
            async with session.request(method, url) as response:
                return response



from unittest import TestCase

from symbiotic.actions import Action
from symbiotic.schedule import Schedule


class Test_Action_Unit(TestCase):

    def test_create_action_without_args(self):
        def mock_callback():
            pass

        action = Action(mock_callback)
        self.assertIsNotNone(action._callback)
        self.assertEqual((), action._callback.args)
        self.assertEqual({}, action._callback.keywords)

    def test_create_action_with_args(self):
        def mock_callback(**kwargs):
            return kwargs.pop('saywhat')

        action = Action(mock_callback, saywhat='what')
        response = action()

        self.assertEqual(action._callback.keywords, {'saywhat': 'what'})
        self.assertEqual('what', response)

    def test_create_action_with_schedule(self):
        def callback():
            pass
        action = Action(callback)
        schedule = Schedule().every_day().at('12:15')
        action.set_schedule(schedule)



from dependency_injector import containers, providers
from gpiozero.pins.pigpio import PiGPIOFactory

from .devices import LightBulb
from .event_bus import EventBus
from .sensors import GPIOMotionSensor
from .services import IFTTT


class ServiceContainer(containers.DeclarativeContainer):
    config = providers.Configuration()

    # It's not possible to pass the root config 'config.services'
    # because, if the configuration does not contain the service's
    # config (e.g. IFTTT), any object trying to call that service will
    # throw "AttributeError: 'NoneType' object has no attribute 'get'"
    IFTTT = providers.Singleton(IFTTT, config=config.IFTTT)


class DeviceContainer(containers.DeclarativeContainer):
    config = providers.Configuration()

    light_bulb: LightBulb = providers.Factory(LightBulb)


class SensorContainer(containers.DeclarativeContainer):
    config = providers.Configuration()

    event_bus = providers.Dependency(instance_of=EventBus)

    pin_factory = providers.FactoryAggregate(
        pigpio=providers.Factory(
            PiGPIOFactory,
            host=config.pigpio.host,
            port=config.pigpio.port
        ),
    )

    gpio_motion_sensor: GPIOMotionSensor = providers.Factory(
        GPIOMotionSensor,
        event_bus=event_bus
    )


class Container(containers.DeclarativeContainer):
    config = providers.Configuration()

    event_bus = providers.Dependency(instance_of=EventBus)

    devices: DeviceContainer = providers.Container(
        DeviceContainer,
        config=config.devices,
    )

    sensors: SensorContainer = providers.Container(
        SensorContainer,
        config=config.sensors,
        event_bus=event_bus,
    )

    services: ServiceContainer = providers.Container(
        ServiceContainer,
        config=config.services,
    )



from abc import ABC, abstractmethod

'Create a high quality module of reusable and extendable data validators'


# Original source:
# https://github.com/rhettinger/modernpython/blob/master/validators.py

class Validator(ABC):

    def __set_name__(self, owner, name):
        self.private_name = f'_{name}'

    def __get__(self, obj, objtype=None):
        return getattr(obj, self.private_name)

    def __set__(self, obj, value):
        self.validate(value)
        setattr(obj, self.private_name, value)

    @abstractmethod
    def validate(self, value):
        pass


class OneOf(Validator):

    def __init__(self, *options):
        self.options = set(options)

    def validate(self, value):
        if value not in self.options:
            raise ValueError(f'{value!r} not a valid option. Should be one of: {self.options}')


class String(Validator):

    def __init__(self, minsize=0, maxsize=None, predicate=None):
        self.minsize = minsize
        self.maxsize = maxsize
        self.predicate = predicate

    def validate(self, value):
        if not isinstance(value, str):
            raise TypeError('Expected a str')
        if len(value) < self.minsize:
            raise ValueError(f'String is too short. Must be at least {self.minsize} long.')
        if self.maxsize is not None and len(value) > self.maxsize:
            raise ValueError(f'String is too long. Must be no bigger than {self.maxsize} long.')
        if self.predicate is not None and not self.predicate(value):
            raise ValueError(f'Expected {value} to be true for {self.predicate !r}')


class Number(Validator):

    def __init__(self, minvalue=None, maxvalue=None):
        self.minvalue = minvalue
        self.maxvalue = maxvalue

    def validate(self, value):
        if not isinstance(value, (int, float)):
            raise TypeError('Expected an int or float')
        if self.minvalue is not None and value < self.minvalue:
            raise ValueError(f'{value} is too small. Must be at least {self.minvalue}.')
        if self.maxvalue is not None and value > self.maxvalue:
            raise ValueError(f'{value} is too big. Must be no more than {self.maxvalue}.')



from unittest import TestCase

from symbiotic.app import Symbiotic


class Test_Application_Unit(TestCase):

    def test_init(self):
        app = Symbiotic()
        self.assertIsNotNone(app)
        self.assertIsNotNone(app.container)
        self.assertIsNotNone(app.container.devices)
        self.assertIsNotNone(app.container.services)



import re
from abc import ABC, abstractmethod
from typing import Union

from pydantic import BaseModel, validator


class Parameters(BaseModel, ABC):

    class Config:
        allow_mutation = False

    @classmethod
    def create(cls, *args, **kwargs):
        return cls(*args, **kwargs)

    @abstractmethod
    def ifttt(self) -> dict:
        pass


class LightBulbParameters(Parameters):

    color: str = '#ffffff'
    brightness: int = 100
    transition_duration: Union[int, str] = 1000

    def ifttt(self) -> dict:
        return {
            'value1': self.brightness,
            'value2': self.color,
            'value3': self.transition_duration
        }

    @validator('color')
    def validate_color(cls, color) -> str:

        # 6-char hex string #FE10EE
        if type(color) is str:
            if re.match(r'#[a-fA-F0-9]{6}$', color):
                return color

        # RGB tuple (255, 255, 100)
        if type(color) is tuple:
            return '#%02x%02x%02x' % (color[0], color[1], color[2])

        raise TypeError(
            f'Color must be either a hex string or a tuple of RGB values, got {color}')

    @validator('brightness')
    def validate_brightness(cls, value: int) -> int:
        if value < 0 or value > 100:
            raise ValueError(
                f'Brightness must be in the range 0-100, got {value}')
        return value

    @staticmethod
    def _convert_transition_duration(duration: str) -> int:
        value = duration.rstrip('smh')
        unit = duration[len(value):]

        if not value.isnumeric and not int(value):
            raise ValueError('Invalid type for transition duration value.')

        unit = unit.lower()
        value = int(value)
        if unit == 'm':
            value = value * 60
        elif unit == 'h':
            value = value * 3600
        elif unit != 's' and unit != '':
            raise ValueError('Invalid type for transition duration unit.')

        return value

    @validator('transition_duration')
    def validate_transition_duration(cls, value: Union[int, str] = 0) -> 'LightBulbParameters':
        if type(value) is str:
            value = cls._convert_transition_duration(value)

        if value < 0:
            raise ValueError('Transition duration must be a positive number')

        # value in ms
        return value * 1000



import setuptools

with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

setuptools.setup(
    name='symbiotic',
    version='1.0.0-alpha3',
    author='Stefano Frazzetto',
    author_email='stefano+symbiotic@hey.com',
    license='Apache-2.0',
    description='Connect your smart devices and sensors to create complex, schedulable actions.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/StefanoFrazzetto/symbiotic',
    packages=setuptools.find_packages(),
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Topic :: Home Automation',
        'Topic :: Software Development :: Libraries',
        'Topic :: System :: Hardware',
    ],
    keywords=[
        'automation',
        'dependency-injection',
        'event-bus',
        'iot',
        'schedule',
        'security',
        'smart-devices',
    ],
    install_requires=[
        'dependency-injector>=4.11',
        'event-bus',
        'gpiozero',
        'pigpio',
        'pydantic',
        'schema',
        'requests'
    ],
    extras_require={
        'dev': [
            'autopep8',
            'flake8',
        ],
        'test': [
            'pytest',
            'pytest-mock',
            'freezegun',
        ],
        'yaml': ['PyYAML>=5.4']
    },
    project_urls={
        'Bug Reports': 'https://github.com/StefanoFrazzetto/symbiotic/issues',
        'Source': 'https://github.com/StefanoFrazzetto/symbiotic',
    },
)



from datetime import datetime
from unittest import TestCase

import pytest
from freezegun import freeze_time

from symbiotic.schedule import Day, Schedule, ScheduleConfigurationError, Instant


class Test_Schedule_DayTime_Unit(TestCase):

    def test_schedule_is_valid(self):
        schedule = Schedule().every(Day.MONDAY).at('08:00')
        self.assertTrue(schedule.is_valid())

    def test_schedule_is_not_valid_missing_time(self):
        schedule = Schedule().every(Day.MONDAY)
        self.assertFalse(schedule.is_valid())

    def test_schedule_is_not_valid_missing_days(self):
        schedule = Schedule().at('07:35')
        self.assertFalse(schedule.is_valid())


class Test_Schedule_Days_Unit(TestCase):

    @freeze_time("2021-02-17")  # Wednesday
    def test_schedule_one_day(self):
        schedule = Schedule().every(Day.WEDNESDAY).at('06:15')
        self.assertTrue(schedule.is_active_today())

    def test_schedule_not_valid_missing_time(self):
        schedule = Schedule().weekends()
        self.assertFalse(schedule.is_valid())

    def test_empty_schedule(self):
        schedule = Schedule()
        with pytest.raises(ScheduleConfigurationError):
            schedule.is_active_today()

    def test_schedule_between_days_exception(self):
        schedule = Schedule()
        with pytest.raises(ScheduleConfigurationError):
            schedule.between()

    def test_schedule_between_two_days(self):
        schedule = Schedule().between(Day.TUESDAY, Day.THURSDAY)
        expected_days = {Day.TUESDAY, Day.WEDNESDAY, Day.THURSDAY}
        self.assertEqual(expected_days, schedule.days)

    def test_schedule_between_two_days_with_rollover(self):
        schedule = Schedule().between(Day.SATURDAY, Day.TUESDAY)
        expected_days = {Day.SATURDAY, Day.SUNDAY, Day.MONDAY, Day.TUESDAY}
        self.assertEqual(expected_days, schedule.days)

    def test_schedule_between_and_exclude_one_day(self):
        schedule = Schedule().between(Day.SATURDAY, Day.TUESDAY).exclude(Day.MONDAY)
        expected_days = {Day.SATURDAY, Day.SUNDAY, Day.TUESDAY}
        self.assertEqual(expected_days, schedule.days)

    def test_schedule_between_and_exclude_day_not_in_schedule(self):
        schedule = Schedule().between(Day.SATURDAY, Day.TUESDAY).exclude(Day.THURSDAY)
        expected_days = {Day.SATURDAY, Day.SUNDAY, Day.MONDAY, Day.TUESDAY}
        self.assertEqual(expected_days, schedule.days)

    def test_schedule_between_same_day(self):
        schedule = Schedule().between(Day.SATURDAY, Day.SATURDAY)
        expected_days = {Day.SATURDAY, Day.SUNDAY, Day.MONDAY, Day.TUESDAY, Day.WEDNESDAY, Day.THURSDAY, Day.FRIDAY}
        self.assertEqual(expected_days, schedule.days)

    def test_schedule_weekdays(self):
        schedule = Schedule().weekdays()
        expected_days = {day for day in Day} - {Day.SATURDAY, Day.SUNDAY}  # all - weekend
        self.assertEqual(expected_days, schedule.days)

    def test_schedule_weekends(self):
        schedule = Schedule().weekends()
        expected_days = {Day.SATURDAY, Day.SUNDAY}  # all - weekend
        self.assertEqual(expected_days, schedule.days)


class Test_Schedule_Time_Unit(TestCase):

    def test_schedule_valid_time_hms(self):
        schedule = Schedule().at('13:15:45')
        self.assertEqual(13, schedule.time.hour)
        self.assertEqual(15, schedule.time.minute)
        self.assertEqual(45, schedule.time.second)

    def test_schedule_valid_time_hm(self):
        schedule = Schedule().at('14:55')
        self.assertEqual(14, schedule.time.hour)
        self.assertEqual(55, schedule.time.minute)

    def test_schedule_valid_time_h(self):
        schedule = Schedule().at('19')
        self.assertEqual(19, schedule.time.hour)

    @freeze_time("2021-02-19 17:58:00")  # Friday - two minutes before scheduled time
    def test_next_datetime_today(self):
        day = Day.FRIDAY
        scheduled_time = datetime(2021, 2, 19, 18, 0, 0).time()
        instant = Instant(day, scheduled_time)

        next_datetime = instant.next_datetime()
        expected_datetime = datetime(2021, 2, 19, 18, 0, 0)

        self.assertEqual(expected_datetime, next_datetime)

    @freeze_time("2021-02-19 18:00:01")  # Friday - one second after scheduled time
    def test_next_datetime_next_week(self):
        day = Day.FRIDAY
        scheduled_time = datetime(2021, 2, 19, 18, 0, 0).time()
        instant = Instant(day, scheduled_time)

        next_datetime = instant.next_datetime()
        expected_datetime = datetime(2021, 2, 26, 18, 0, 0)  # one week after

        self.assertEqual(expected_datetime, next_datetime)



"""
Colors defined as Enum

The colours have been sourced from:
https://stackoverflow.com/questions/22408237/named-colors-in-matplotlib
"""

# flake8: noqa

from enum import Enum


class Colour(Enum):

    def __repr__(self):
        return f'{self.name}: {self.value}'

    def __str__(self) -> str:
        return self.name

    @property
    def hex(self) -> str:
        return self.value

    @property
    def rgb(self) -> tuple:
        value = self.value.lstrip('#')
        return tuple(int(value[i:i + 2], 16) for i in (0, 2, 4))

    ALICE_BLUE =             '#F0F8FF'
    ANTIQUE_WHITE =          '#FAEBD7'
    AQUA =                   '#00FFFF'
    AQUAMARINE =             '#7FFFD4'
    AZURE =                  '#F0FFFF'
    BEIGE =                  '#F5F5DC'
    BISQUE =                 '#FFE4C4'
    BLACK =                  '#000000'
    BLANCHEDALMOND =         '#FFEBCD'
    BLUE =                   '#0000FF'
    BLUEVIOLET =             '#8A2BE2'
    BROWN =                  '#A52A2A'
    BURLYWOOD =              '#DEB887'
    CADET_BLUE =             '#5F9EA0'
    CHARTREUSE =             '#7FFF00'
    CHOCOLATE =              '#D2691E'
    CORAL =                  '#FF7F50'
    CORNFLOWER_BLUE =        '#6495ED'
    CORNSILK =               '#FFF8DC'
    CRIMSON =                '#DC143C'
    CYAN =                   '#00FFFF'
    DARK_BLUE =              '#00008B'
    DARK_CYAN =              '#008B8B'
    DARK_GOLDENROD =         '#B8860B'
    DARK_GRAY =              '#A9A9A9'
    DARK_GREEN =             '#006400'
    DARK_KHAKI =             '#BDB76B'
    DARK_MAGENTA =           '#8B008B'
    DARK_OLIVEGREEN =        '#556B2F'
    DARK_ORANGE =            '#FF8C00'
    DARK_ORCHID =            '#9932CC'
    DARK_RED =               '#8B0000'
    DARK_SALMON =            '#E9967A'
    DARK_SEAGREEN =          '#8FBC8F'
    DARK_SLATEBLUE =         '#483D8B'
    DARK_SLATEGRAY =         '#2F4F4F'
    DARK_TURQUOISE =         '#00CED1'
    DARK_VIOLET =            '#9400D3'
    DEEP_PINK =              '#FF1493'
    DEEP_SKYBLUE =           '#00BFFF'
    DIMGRAY =                '#696969'
    DODGERBLUE =             '#1E90FF'
    FIREBRICK =              '#B22222'
    FLORAL_WHITE =           '#FFFAF0'
    FOREST_GREEN =           '#228B22'
    FUCHSIA =                '#FF00FF'
    GAINSBORO =              '#DCDCDC'
    GHOST_WHITE =            '#F8F8FF'
    GOLD =                   '#FFD700'
    GOLDENROD =              '#DAA520'
    GRAY =                   '#808080'
    GREEN =                  '#008000'
    GREEN_YELLOW =           '#ADFF2F'
    HONEYDEW =               '#F0FFF0'
    HOT_PINK =               '#FF69B4'
    INDIAN_RED =             '#CD5C5C'
    INDIGO =                 '#4B0082'
    IVORY =                  '#FFFFF0'
    KHAKI =                  '#F0E68C'
    LAVENDER =               '#E6E6FA'
    LAVENDER_BLUSH =         '#FFF0F5'
    LAWNGREEN =              '#7CFC00'
    LEMONCHIFFON =           '#FFFACD'
    LIGHT_BLUE =             '#ADD8E6'
    LIGHT_CORAL =            '#F08080'
    LIGHT_CYAN =             '#E0FFFF'
    LIGHT_GOLDENROD_YELLOW = '#FAFAD2'
    LIGHT_GREEN =            '#90EE90'
    LIGHT_GRAY =             '#D3D3D3'
    LIGHT_PINK =             '#FFB6C1'
    LIGHT_SALMON =           '#FFA07A'
    LIGHT_SEAGREEN =         '#20B2AA'
    LIGHT_SKYBLUE =          '#87CEFA'
    LIGHT_SLATEGRAY =        '#778899'
    LIGHT_STEELBLUE =        '#B0C4DE'
    LIGHT_YELLOW =           '#FFFFE0'
    LIME =                   '#00FF00'
    LIMEGREEN =              '#32CD32'
    LINEN =                  '#FAF0E6'
    MAGENTA =                '#FF00FF'
    MAROON =                 '#800000'
    MEDIUM_AQUAMARINE =      '#66CDAA'
    MEDIUM_BLUE =            '#0000CD'
    MEDIUM_ORCHID =          '#BA55D3'
    MEDIUM_PURPLE =          '#9370DB'
    MEDIUM_SEAGREEN =        '#3CB371'
    MEDIUM_SLATEBLUE =       '#7B68EE'
    MEDIUM_SPRINGGREEN =     '#00FA9A'
    MEDIUM_TURQUOISE =       '#48D1CC'
    MEDIUM_VIOLETRED =       '#C71585'
    MIDNIGHT_BLUE =          '#191970'
    MINTCREAM =              '#F5FFFA'
    MISTYROSE =              '#FFE4E1'
    MOCCASIN =               '#FFE4B5'
    NAVAJOWHITE =            '#FFDEAD'
    NAVY =                   '#000080'
    OLDLACE =                '#FDF5E6'
    OLIVE =                  '#808000'
    OLIVEDRAB =              '#6B8E23'
    ORANGE =                 '#FFA500'
    ORANGERED =              '#FF4500'
    ORCHID =                 '#DA70D6'
    PALE_GOLDENROD =         '#EEE8AA'
    PALE_GREEN =             '#98FB98'
    PALE_TURQUOISE =         '#AFEEEE'
    PALE_VIOLETRED =         '#DB7093'
    PAPAYA_WHIP =            '#FFEFD5'
    PEACHPUFF =              '#FFDAB9'
    PERU =                   '#CD853F'
    PINK =                   '#FFC0CB'
    PLUM =                   '#DDA0DD'
    POWDER_BLUE =            '#B0E0E6'
    PURPLE =                 '#800080'
    RED =                    '#FF0000'
    ROSYBROWN =              '#BC8F8F'
    ROYAL_BLUE =             '#4169E1'
    SADDLE_BROWN =           '#8B4513'
    SALMON =                 '#FA8072'
    SANDY_BROWN =            '#FAA460'
    SEA_GREEN =              '#2E8B57'
    SEA_SHELL =              '#FFF5EE'
    SIENNA =                 '#A0522D'
    SILVER =                 '#C0C0C0'
    SKYBLUE =                '#87CEEB'
    SLATE_BLUE =             '#6A5ACD'
    SLATE_GRAY =             '#708090'
    SNOW =                   '#FFFAFA'
    SPRING_GREEN =           '#00FF7F'
    STEEL_BLUE =             '#4682B4'
    TAN =                    '#D2B48C'
    TEAL =                   '#008080'
    THISTLE =                '#D8BFD8'
    TOMATO =                 '#FF6347'
    TURQUOISE =              '#40E0D0'
    VIOLET =                 '#EE82EE'
    WHEAT =                  '#F5DEB3'
    WHITE =                  '#FFFFFF'
    WHITE_SMOKE =            '#F5F5F5'
    YELLOW =                 '#FFFF00'
    YELLOW_GREEN =           '#9ACD32'



from abc import ABC, abstractmethod
from enum import Enum

from .exceptions import ConfigurationError
from .parameters import LightBulbParameters, Parameters
from .services import ServiceResponse, BaseService


class State(Enum):
    ON = 'on'
    OFF = 'off'

    def __str__(self):
        return self.value


class SmartDevice(ABC):
    """
    SmartDevice encapsulates the methods to control any smart device.

    Devices are controlled through a BaseService; using the facade pattern
    here allows to add more services in the future without refactoring,
    improves readability, and reduces code coupling.
    """

    "Map device physical states to IFTTT service_event names."
    state_event_mapping: dict = {
        State.ON: 'bedroom_light_color',
        State.OFF: 'switch_off'
    }

    def __init__(self, name: str, *args, **kwargs) -> None:
        self.name: str = name
        self.service: BaseService = kwargs.pop('service', None)
        self._state: State = kwargs.pop('state', None)
        self._parameters: Parameters = self.default_parameters
        super().__init__(*args, **kwargs)

    @property
    @abstractmethod
    def default_parameters(self) -> Parameters:
        raise NotImplementedError(
            "The subclass must override 'parameters'.")

    @property
    def parameters(self) -> Parameters:
        return self._parameters

    @parameters.setter
    def parameters(self, params):
        if not isinstance(params, Parameters):
            raise TypeError(f'Wrong type for parameters, got {params}')

        self._parameters = params

    @property
    def state(self) -> State:
        return self._state

    @state.setter
    def state(self, state: 'State') -> None:
        self._state = state

    def _change_state(self, state: 'State', **kwargs) -> ServiceResponse:
        if not self.service:
            raise ConfigurationError('You need to add a service to this device')

        parameters = self.parameters.create(**kwargs)
        response = self.service.trigger(
            event_name=SmartDevice.state_event_mapping[state],
            parameters=parameters
        )

        # update device state and parameters
        if response.success:
            self.state = state
            self.parameters = parameters

        return response


class LightBulb(SmartDevice):

    def __init__(self, name: str, *args, **kwargs):
        super().__init__(name, *args, **kwargs)

    @property
    def default_parameters(self):
        return LightBulbParameters()

    def turn_on(self, **params) -> ServiceResponse:
        return self._change_state(State.ON, **params)

    def turn_off(self, **params) -> ServiceResponse:
        return self._change_state(State.OFF, **params)
